/*
-----------------------------------------------------
   Application Name: "DIMG"
Application Version: 0.1.0
          File Name: Dimg.cpp
       File Version: 0.1.0
         Start Date: 27/06/2008
          Author(s): David Goodwin
          Copyright: (C) Copyright David Goodwin, 2008
            License: GNU General Public License
   File Description: DIMG Class Implementation
-----------------------------------------------------
	This file is part of "DIMG".

    This is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This software is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this software; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

	File Notes:
	------------------
*/



// ------------------
// Header files:
// ------------------
#include "Dimg.h"

// ------------------
// Win32 Stuff:
// ------------------
#ifdef Q_OS_WIN32
// This stuff is to support autodetecting removable media drives on Windows systems.
#include "windows.h"
static const QString win32_drives[] = {"A", "B", "C", "D", "E", "F", 
									   "G", "H", "I", "J", "K", "L", 
									   "M", "N", "O", "P", "Q", "R", 
									   "S", "T", "U", "V", "W", "X", 
									   "Y", "Z"};
static const int win32_drive_count = 26;
#endif

// ------------------
// Constructor for download_manager class
Dimg::Dimg(QApplication *app) : QMainWindow(0)
{
	// Remember the pointer to the application
	pApp = app;
	// Then setup the Ui
	ui.setupUi(this);

	const QStringList fds = getFloppyDrives();

	// Create the list of floppy drives
	for (int i = 0; i < fds.count(); i++) {
		ui.format_floppy_list->addItem(fds.at(i));
		ui.copy_src_floppy->addItem(fds.at(i));
		ui.copy_dest_floppy->addItem(fds.at(i));
	}

	// Create the list of disk formats
	ui.format_formatselect->addItems(QLibDsk::list_format_descriptions());
	// Set the default disk format to IBM 1.44MB
	ui.format_formatselect->setCurrentIndex(10);

	// connect up signals and slots. Note that some signals an slots have also 
	// been setup in the UI file (mostly to deal with the radio buttons and
	// their associated text or combination widgets)
	connect(ui.info_go_button,SIGNAL(clicked()),this,SLOT(getInfo()));
	connect(ui.format_go, SIGNAL(clicked()),this,SLOT(formatDisk()));
	connect(ui.copy_go, SIGNAL(clicked()),this,SLOT(copyDisk()));
	connect(ui.aboutQt, SIGNAL(clicked()),app, SLOT(aboutQt()));

	// Browse buttons
	connect(ui.format_image_browse, SIGNAL(clicked()),this,SLOT(browseFormatImage()));
	connect(ui.copy_src_browse, SIGNAL(clicked()),this,SLOT(browseCopySrcImage()));
	connect(ui.copy_dest_browse, SIGNAL(clicked()),this,SLOT(browseCopyDestImage()));

	// Setup the table on the about tab:
	QStringList data;
	data << "Qt Version";
	data << QString::fromAscii(qVersion());
	ui.tvAboutProperties->addTopLevelItem(new QTreeWidgetItem(data,0));
	data.clear();

	data << "LibDsk Version";
	data << QString::fromAscii(LIBDSK_VERSION);
	ui.tvAboutProperties->addTopLevelItem(new QTreeWidgetItem(data,0));
	data.clear();

	// Include some more detailed build information if the compiler was GNU GCC
#ifdef __GNUC__
	data << "Compiler";
	data << QString::fromAscii("GNU GCC");
	ui.tvAboutProperties->addTopLevelItem(new QTreeWidgetItem(data,0));
	data.clear();

	data << "GCC Version";
	data << QString::fromAscii(__VERSION__);
	ui.tvAboutProperties->addTopLevelItem(new QTreeWidgetItem(data,0));
	data.clear();

	data << "Build Date";
	data << QString::fromAscii(__DATE__);
	ui.tvAboutProperties->addTopLevelItem(new QTreeWidgetItem(data,0));
	data.clear();
#endif

	setWindowTitle(application_name + " version " + versionString());

	// Show the user interface.
	show();
}

// This function gets disk image information.
void Dimg::getInfo() {
	// First up, get the file or image name:
	QString filename = ui.info_filename->text();
	
	dsk_err_t err;

	QLibDsk dsk;
	err = dsk.open(filename);

	if (err) {
		QString title = "Error";
		if (err == DSK_ERR_BADPTR   || err == DSK_ERR_DIVZERO ||
			err == DSK_ERR_BADPARM   ) title = "Error: Program bug?";
		QString errstring = QLibDsk::strerror(err);

		QMessageBox::critical(this, title, errstring);
	}

	QString compdesc = dsk.compdesc();
	QString driverdesc = dsk.drvdesc();
	QStringList data;
	DSK_GEOMETRY dg;
	dsk.getgeom(&dg);
	unsigned char drv_status;

	ui.info_list->clear();

	data << "Driver";
	data << driverdesc;
	ui.info_list->addTopLevelItem(new QTreeWidgetItem(data,0));

	data.clear();
	data << "Compression";
	data << compdesc;
	ui.info_list->addTopLevelItem(new QTreeWidgetItem(data,0));

	QString str = "";

	data.clear();
	data << "Sidedness";
	data << str.sprintf("%d",dg.dg_sidedness);
	ui.info_list->addTopLevelItem(new QTreeWidgetItem(data,0));
	str.clear();

	data.clear();
	data << "Cylinders";
	data << str.sprintf("%d",dg.dg_cylinders);
	ui.info_list->addTopLevelItem(new QTreeWidgetItem(data,0));
	str.clear();

	data.clear();
	data << "Heads";
	data << str.sprintf("%d",dg.dg_heads);
	ui.info_list->addTopLevelItem(new QTreeWidgetItem(data,0));
	str.clear();

	data.clear();
	data << "Sectors";
	data << str.sprintf("%d",dg.dg_sectors);
	ui.info_list->addTopLevelItem(new QTreeWidgetItem(data,0));
	str.clear();

	data.clear();
	data << "Sector Size";
	data << str.sprintf("%d",dg.dg_secsize);
	ui.info_list->addTopLevelItem(new QTreeWidgetItem(data,0));
	str.clear();

	data.clear();
	data << "First Sector";
	data << str.sprintf("%d",dg.dg_secbase);
	ui.info_list->addTopLevelItem(new QTreeWidgetItem(data,0));
	str.clear();

	data.clear();
	data << "Data Rate";
	data << str.sprintf("%d",dg.dg_datarate);
	ui.info_list->addTopLevelItem(new QTreeWidgetItem(data,0));
	str.clear();

	data.clear();
	data << "Record Mode";
	if (dg.dg_fm == 0)
		data << "MFM";
	else
		data << "FM";
	ui.info_list->addTopLevelItem(new QTreeWidgetItem(data,0));
	str.clear();

	data.clear();
	data << "R/W Gap";
	data << str.sprintf("0x%02x",dg.dg_rwgap);
	ui.info_list->addTopLevelItem(new QTreeWidgetItem(data,0));
	str.clear();

	data.clear();
	data << "Format Gap";
	data << str.sprintf("0x%02x",dg.dg_fmtgap);
	ui.info_list->addTopLevelItem(new QTreeWidgetItem(data,0));
	str.clear();

	data.clear();
	err = dsk.drive_status(&dg,0,&drv_status);
	if (err) {
		QString title = "Error";
		if (err == DSK_ERR_BADPTR   || err == DSK_ERR_DIVZERO ||
			err == DSK_ERR_BADPARM   ) title = "Error: Program bug?";
		QString errstring = QLibDsk::strerror(err);

		QMessageBox::critical(this, title, errstring);
	}
	data << "Drive Status";
	if (!err)
		data << str.sprintf("0x%02x",drv_status);
	else
		data << strerror(err);
	ui.info_list->addTopLevelItem(new QTreeWidgetItem(data,0));
	str.clear();
	
	data.clear();
	QString comment;
	err = dsk.get_comment(&comment);
	if (err == DSK_ERR_OK && comment != NULL) {
		data << "Comment";
		data << comment;
		ui.info_list->addTopLevelItem(new QTreeWidgetItem(data,0));		
	}
	if (err) {
		QString title = "Error";
		if (err == DSK_ERR_BADPTR   || err == DSK_ERR_DIVZERO ||
			err == DSK_ERR_BADPARM   ) title = "Error: Program bug?";
		QString errstring = QLibDsk::strerror(err);

		QMessageBox::critical(this, title, errstring);
	}

	dsk.close();
}

void Dimg::formatDisk(){
	// Figure out if we are formatting a floppy disk or an image file:
	QString filename = "";
	dsk_format_t format;
	bool isFloppy = false;

	if (ui.format_opt_floppy->isChecked()) {
		// We are formatting a floppy disk
		filename = ui.format_floppy_list->currentText();
		isFloppy = true;
	}
	else {
		// We are formatting an image file
		filename = ui.format_filename->text();
	}

	// Get the format:
	format = (dsk_format_t)ui.format_formatselect->currentIndex();

	// Set the format going while updating progress information
	QLibDsk dsk;
	DSK_GEOMETRY dg;
	dsk_err_t err;
	dsk_pcyl_t cyl;
	dsk_phead_t head;
	dsk_cchar_t fdesc;
	bool stopped = false;

	// First up, create the disk:
	
	QString type = "raw";
	if (isFloppy) type = "floppy";
	// Create the disk image or open the floppy drive
	err = dsk.creat(filename, type.toAscii().constData(), NULL);
	if (err) {
		QMessageBox::critical(this, "Error", QLibDsk::strerror(err));
		return;
	}
	// Set the number of retries
	err = dsk.set_retry(5);
	if (err) {
		QMessageBox::critical(this, "Error", QLibDsk::strerror(err));
		return;
	}
	// Select the disk format
	err = dsk.stdformat(&dg,format,NULL,&fdesc);

	if (err) {
		QMessageBox::critical(this, "Error", QLibDsk::strerror(err));
		return;
	}

	if (!err) {
		// max is one larger than it needs to be so that the dialog doesnt disappear before the first sector is
		// written.
		unsigned max = (dg.dg_heads * dg.dg_cylinders) +1;
		
		// Setup the progress dialog
		QString format_string = "Formatting disk " + filename + " as " + fdesc;
		QProgressDialog progress("Formatting disk","Abort Format",0,max,this);
		progress.setWindowTitle("Formatting disk");
		progress.setLabelText(format_string);
		progress.setWindowModality(Qt::WindowModal);


		QString cl_count;
		QString hd_count;

		cl_count.setNum(dg.dg_cylinders);
		hd_count.setNum(dg.dg_heads);
		// Set the functions to be called to report status information
		//dsk.reportfunc_set(report, report_end);

		// Do the format
		for (cyl = 0; cyl < dg.dg_cylinders; cyl++) {
			for (head = 0; head < dg.dg_heads; head++) {
				// Update progress information
				QString cl;
				QString hd;

				cl.setNum(cyl+1);
				hd.setNum(head+1);
				progress.setLabelText("Formatting cylinder " + cl + "/" 
					+ cl_count + " Head " + hd + "/" 
					+ hd_count);
				progress.setValue(cyl*dg.dg_heads + head);

				// Check to see if the user canceled
				QApplication::processEvents();
				if (progress.wasCanceled())
					stopped = true;
				if (stopped) break;
				
				err = dsk.apform(&dg,cyl,head,0xE5);
				if (err) break;
			}
			if (stopped || err) break;
		}
		// Create a disc spec on the first sector if it is a recognised format
		if (!stopped && !err) {
			unsigned char bootsec[512];
			unsigned int do_bootsec = 0;

			progress.setLabelText("Writing first sector");
			memset(bootsec,0xE5,sizeof(bootsec));
			switch(format) {
				case FMT_180K:
						memcpy(bootsec, spec180, 10);
						do_bootsec = 1;
						break;
					case FMT_CPCSYS:
						memcpy(bootsec, spec169, 10);
						do_bootsec = 1;
						break;
					case FMT_CPCDATA:
						break;
					case FMT_200K:
						memcpy(bootsec, spec200, 10);
						do_bootsec = 1;
						break;
					case FMT_720K:
						memcpy(bootsec, spec720, 10);
						do_bootsec = 1;
						break;
					case FMT_800K:
						memcpy(bootsec, spec800, 10);
						do_bootsec = 1;
						break;
					default:
						break;
			}
			if (do_bootsec) err = dsk.lwrite(&dg, bootsec, 0);
			progress.setValue(max+1);
		}
		dsk.close();
		if (err) {
			QMessageBox::critical(0,"Format Error","Format Error: " + dsk.strerror(err),QMessageBox::Ok);
			return;
		}
		else if (stopped) {
			QMessageBox::warning(0,"Format Canceled","The format was cancelled by the user. The disk is not completely formateed - do not use it.",QMessageBox::Ok);
		}
		else {
			QMessageBox::information(0,"Format Complete","Disk Format Complete.",QMessageBox::Ok);
		}
	}
	if (err) {
		QMessageBox::critical(this, "Error", QLibDsk::strerror(err));
		return;
	}
}

void Dimg::copyDisk() {

	// Some stuff we will need later.
	char *buf = NULL;			// A buffer
	bool noFormat = false;		// Set to TRUE to disable formatting
	int last = -1, first = -1;	// Change these to do a partial copy

	// We need two disks: a source and a target
	QLibDsk source_disk, target_disk;		// The source and target disks

	// Various stuff we need while copying the disk
	DSK_GEOMETRY dg;					// Disk geometry of the two disks
	dsk_err_t err = DSK_ERR_NOTIMPL;	// What ever error may occur goes in here
	dsk_pcyl_t cyl;			// Cylinder
	dsk_phead_t head;		// Head
	dsk_psect_t sec;		// Sector
	dsk_cchar_t fdesc;		// 
	bool stopped = false;	// Set to TRUE to stop the copy

	// This stores the filenames we are using for the source and target and
	// what they are (floppy drive or image file)
	bool isSourceFloppy = false, isTargetFloppy = false;
	QString source_filename, target_filename;

	// Get the filename and type for the source disk
	if (ui.copy_opt_src_floppy->isChecked()) {
		// We are copying from a floppy disk
		source_filename = ui.copy_src_floppy->currentText();
		isSourceFloppy = true;
	}
	else {
		// We are copying from an image file
		source_filename = ui.copy_src_filename->text();
	}	

	// Get the filename and type for the target disk
	if (ui.copy_opt_dest_floppy->isChecked()) {
		// We are copying to a floppy disk
		target_filename = ui.copy_dest_floppy->currentText();
		isTargetFloppy = true;
	}
	else {
		// We are copying to an image file
		target_filename = ui.copy_dest_filename->text();
	}	

	// Now open the source disk
	// ------------------------
	err = source_disk.open(source_filename);
	if (err) {
		QMessageBox::critical(this, "Error", QLibDsk::strerror(err));
		return;
	}

	// Set the number of retries. 5 seems like a good number.
	err = source_disk.set_retry(5);
	if (err) {
		QMessageBox::critical(this, "Error", QLibDsk::strerror(err));
		return;
	}


	// Now create the target disk
	// --------------------------

	// We cant autodetect the target type so we must specify it.
	// If the user selected a floppy drive then set that as the type.
	QString type = "raw";
	if (isTargetFloppy) type = "floppy";

	// Create the disk image or open the floppy drive
	err = target_disk.creat(target_filename, type.toAscii().constData(), NULL);
	if (err) {
		QMessageBox::critical(this, "Error", QLibDsk::strerror(err));
		return;
	}

	// Set the number of retries. 5 seems good.
	err = target_disk.set_retry(5);
	if (err) {
		QMessageBox::critical(this, "Error", QLibDsk::strerror(err));
		return;
	}

	// Setup formats
	// -------------

	err = source_disk.getgeom(&dg);
	if (err) {
		QMessageBox::critical(this, "Error", QLibDsk::strerror(err));
		return;
	}

	// We want to copy the _entire_ disk. If you want to do a partial
	// copy, change the values of first and last to those desired at the start
	// of this function.
	if (last == -1) last = dg.dg_cylinders - 1;
	if (first == -1) first = 0;

	buf = (char *)dsk_malloc(dg.dg_secsize);
	if (!buf) err = DSK_ERR_NOMEM;

	if (err) {
		QMessageBox::critical(this, "Error", QLibDsk::strerror(err));
		return;
	}

	unsigned max = dg.dg_heads * (last - first + 1) * dg.dg_sectors;

	// Setup progress dialog
	QString format_string = "Copying from " + source_filename + " to " + target_filename;
	QProgressDialog progress("Copying disk","Abort Copy",0,max,this);
	progress.setWindowTitle("Copying disk");
	progress.setLabelText(format_string);
	progress.setWindowModality(Qt::WindowModal);
	QString qsCyl;
	QString qsCylT;
	QString qsHead;
	QString qsHeads; 
	QString qsEct; // qs = QString, sEct = Sector
	QString qsEcT; // qs = QString, sEcT = Sector, T = total

	// Copy the comment, if there is any.
	QString comment;
	source_disk.get_comment(&comment);
	target_disk.set_comment(comment.toAscii().constData());

	// The file the error occured in.
	QString errorFile = "";

	// now, start copying.
	for (cyl = first; (int)cyl <= last; ++cyl) {
		for (head = 0; head < dg.dg_heads; ++head) {

			// Format the track
			if (err == DSK_ERR_OK && (!noFormat)) {
				errorFile = target_filename;
				target_disk.apform(&dg, cyl, head, 0xE5);
			}

			// If we havent had any errors, loop over the sectors.
			if (!err) for (sec = 0; sec < dg.dg_sectors; ++sec) {
				// Set the progress label. This is complicated! (and could probably be done much simpler
				qsCyl.setNum(cyl +1);
				qsCylT.setNum(dg.dg_cylinders);
				qsHead.setNum(head +1);
				qsHeads.setNum(dg.dg_heads);
				qsEct.setNum(sec+dg.dg_secbase);
				qsEcT.setNum(dg.dg_sectors + dg.dg_secbase - 1);
				progress.setLabelText("Copying cylinder " + qsCyl + "/" 
					+ qsCylT + " Head " + qsHead + "/" 
					+ qsHeads + " Sector " + qsEct + "/"
					+ qsEcT);

				progress.setValue(dg.dg_sectors * (cyl * dg.dg_heads + head) + sec);

				// Check to see if the user canceled
				//QApplication::processEvents();
				if (progress.wasCanceled()) {
					stopped = true;
				}

				errorFile = source_filename;
				err = source_disk.pread(&dg, buf, cyl, head, sec + dg.dg_secbase);

				if (err) break;

				errorFile = target_filename;
				err = target_disk.pwrite(&dg,buf,cyl,head, sec + dg.dg_secbase);
				if (err) break;
			}
			if (err || stopped) break;
		}
		if (err || stopped) break;
	}
	// we dont need to delete the progress thingey
	source_disk.close();
	target_disk.close();

	if (err) {
		if (!errorFile.isEmpty()) {
			// Build the error string
			QString msg = "An Error occured while";

			if (errorFile == source_filename) {
				msg.append(" reading from");
				if (isSourceFloppy) msg.append(" floppy drive ");
				else msg.append(" the image file ");
				msg.append(errorFile);
			}
			else {
				msg.append(" writing to");
				if (isSourceFloppy) msg.append(" floppy drive ");
				else msg.append(" the image file ");
				msg.append(errorFile);
			}
			msg.append(". The error was: " + QLibDsk::strerror(err));
			QMessageBox::critical(this, "Error", msg);
		}
		else QMessageBox::critical(this, "Error", QLibDsk::strerror(err));
		return;
	}
	if (stopped) {
		QMessageBox::warning(this, "Copy canceled by user", "Destination disk or image file is not complete - don't use it.");
	}
	else {
		QMessageBox::information(this, "Copy Complete","Copy completed successfully");
	}

	if (buf) dsk_free(buf);
}


QString Dimg::versionString() {
	QString major; major.setNum(major_version);
	QString minor; minor.setNum(minor_version);
	QString patch; patch.setNum(patch_level);

	QString versionString = major + "." + minor;

	if (patch_level > 0) {
		if (use_standard_version_string)
			versionString.append(".");
		else
			versionString.append("-");
		 versionString.append(patch);
	}

	return  versionString;
}
QStringList Dimg::getFloppyDrives() {
	QStringList qsl;
	for (int i = 0; i < floppy_drive_count; i++)
		qsl.append(QString(floppy_drives[i]));	

#ifdef Q_OS_WIN32
	if (detect_floppy_drives) {
		qsl.clear();
		for (int j = 0; j < win32_drive_count; j++) {
			UINT dt = GetDriveTypeA(QString(win32_drives[j] + ":").toAscii().constData());

			QString msg = "Drive ";
			msg.append(win32_drives[j]);
			msg.append(" is ");

			if (dt == DRIVE_UNKNOWN)
				msg.append("DRIVE_UNKNOWN");
			else if (dt == DRIVE_NO_ROOT_DIR)
				msg.append("DRIVE_NO_ROOT_DIR");
			else if (dt == DRIVE_REMOVABLE) {
				msg.append("DRIVE_REMOVABLE");
				qsl.append(QString(win32_drives[j]) + ":");
			}
			else if (dt == DRIVE_FIXED)
				msg.append("DRIVE_FIXED");
			else if (dt == DRIVE_REMOTE)
				msg.append("DRIVE_REMOTE");
			else if (dt == DRIVE_CDROM)
				msg.append("DRIVE_CDROM");
			else if (dt == DRIVE_RAMDISK)
				msg.append("DRIVE_RAMDISK");
			else 
				msg.append("unknown :(");

			// Debugging stuff
			//QMessageBox::information(NULL, "", msg);
		}
	}
#endif

	return qsl;
}


void Dimg::browseFormatImage() {
	QString fileName = QFileDialog::getSaveFileName(this, "Save Disk Image", NULL, "All Files (*.*)");
	ui.format_filename->setText(fileName);
}

void Dimg::browseCopySrcImage() {
	QString fileName = QFileDialog::getOpenFileName(this, "Open Disk Image", NULL, "All Files (*.*)");
	ui.copy_src_filename->setText(fileName);
} 

void Dimg::browseCopyDestImage() {
	QString fileName = QFileDialog::getSaveFileName(this, "Save Disk Image", NULL, "All Files (*.*)");
	ui.copy_dest_filename->setText(fileName);
}
